#ifndef OPTIMIZERPORTFOLIO_H
#define OPTIMIZERPORTFOLIO_H
#include "ClpSimplex.hpp"
#include <boost/lexical_cast.hpp>
#include <Eigen/Dense>
#include "StOpt/sddp/SDDPCutBase.h"
#include "StOpt/sddp/OptimizerSDDPBase.h"
#include "StOpt/core/utils/constant.h"



/// \class AddConstraintOptimizerPortfolio  OptimizerPortfolio.h
/// Add constraints to the Bellman value with cuts
class AddConstraintOptimizerPortfolio
{

protected:

    /// \brief calculate cuts
    /// \param  p_linCut         cuts stored
    virtual inline 	 Eigen::ArrayXXd calCuts(const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut) const
    {
        return  Eigen::ArrayXXd() ;
    }

public:
    /// \brief  add constraints to Bellman value
    /// \param  p_linCut          cuts stored
    /// \param  p_nbAsset         number of risky assets
    /// \param  p_stateFollowing  state at next date
    /// \param  p_rows            rows for matrix contraints
    /// \param  p_columns         columns for matrix contraints
    /// \param  p_elements        A matrix elements
    /// \param  p_lowBoundConst   lower constraint \f$ lc\f$  on matrix \f$ lc \le A x \f$
    /// \param  p_upperBoundConst upper constraint \f$ uc\f$  on matrix \f$ A x \le uc \f$
    void addConstraints(const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut, const int &p_nbAsset , Eigen::ArrayXi &p_rows,   Eigen::ArrayXi   &p_columns,  Eigen::ArrayXd   &p_elements,
                        Eigen::ArrayXd    &p_lowBoundConst,  Eigen::ArrayXd   &p_upperBoundConst) const
    {
        // get back cuts
        Eigen::ArrayXXd  cuts = calCuts(p_linCut);
        int iBellPos = 3 * p_nbAsset + 1; // offset in variables
        int isizeInit = p_elements.size();
        // below cuts are conditional to the risky assets and the non risky one.
        p_rows.conservativeResize(isizeInit + (p_nbAsset + 2)*cuts.cols());
        p_columns.conservativeResize(isizeInit + (p_nbAsset + 2)*cuts.cols());
        p_elements.conservativeResize(isizeInit + (p_nbAsset + 2)*cuts.cols());
        int ibound = p_lowBoundConst.size();
        p_lowBoundConst.conservativeResize(ibound + cuts.cols());
        p_upperBoundConst.conservativeResize(ibound + cuts.cols());
        // add cuts
        for (int icut = 0 ; icut < cuts.cols() ; ++icut)
        {
            double affineValue = cuts(0, icut);
            int ipos = isizeInit + (p_nbAsset + 2) * icut;
            p_rows(ipos) = ibound + icut;
            p_columns(ipos) = iBellPos;
            p_elements(ipos) = 1;
            for (int isto = 0 ; isto <=  p_nbAsset ; ++isto)
            {
                p_rows(ipos + isto + 1) = ibound + icut;
                p_columns(ipos + isto + 1) = isto;
                double derivStorage =  cuts(1 + isto, icut);
                p_elements(ipos + isto + 1) = -derivStorage;
            }
            p_lowBoundConst(ibound + icut) = affineValue;
            p_upperBoundConst(ibound + icut) = StOpt::infty;
        }
    }
};


/// \class AddConstraintOptimizerPortfolioOpti OptimizePortfolio.h
/// Add constraints to the Bellman value with cuts in optimizer part
class AddConstraintOptimizerPortfolioOpti :  public AddConstraintOptimizerportfolio
{
private:

    /// \brief simulation number
    int m_isim ;

    /// \brief calculate cuts
    /// \param  p_linCut         cuts stored
    virtual inline 	 Eigen::ArrayXXd calCuts(const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut) const
    {
        return  p_linCut->getCutsAssociatedToTheParticle(m_isim);
    }

public :

    /// \brief Constructor
    /// \param p_isim  simulation number used in optimization part
    AddConstraintOptimizerPortfolioopti(const int &p_isim): m_isim(p_isim) {};
};



/// \class AddConstraintOptimizerPortfoliosimu  OPTIMIZERPORTFOLIOMANAGEMENT_H
/// Add constraints to the Bellman value with cuts in simulation part
class AddConstraintOptimizerPortfoliosimu :  public AddConstraintOptimizerPortfolio
{

private:

    /// \brief Uncertainty used in regression
    Eigen::ArrayXd m_alea;

    /// \brief calculate cuts
    /// \param  p_linCut         cuts stored
    virtual inline 	 Eigen::ArrayXXd calCuts(const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut) const
    {
        return  p_linCut->getCutsAssociatedToAParticle(m_alea);
    }

public :

    /// \brief Constructor
    /// \param p_alea   uncertainty obtained during simulation part
    AddConstraintOptimizerPortfoliosimu(const Eigen::ArrayXd   &p_alea): m_alea(p_alea) {};

};




/// Define a test case with N assets for portfolio optimization with SDDP.
/// A portfolio is invested in N risky assets \f$S_i\f$ and a none risky asset.
/// The none risky asset has a zero yield.
/// The risky assets \f$ S^i \f$ has a stochastic yield into two dates \f$ t- dt\d$ and \f$ t \$ :
///    \f$ a^i_t = \mu^i + \sigma^i g^i_t \f$ where the \f$ g^i_t \f$ at a given date \f$ t\f$
/// are centered unit gaussian with correlations matrice \$f \Sigma \f$
/// so that $S^i_t = (1+a^i_t) S^i_{t-dt}$
/// \f$ X_t= (X^1_t,..,X^N,t, ...., X^{3N}_t, X^{3N+1}_t \f$ is the state vector.
///  - \f$ X^i_t,  i \[1,N] \f$ represents the positive amount invested in asset \f$i\f$ at date \f$t\f$ before rebalancing
///  - \f$ X^i_t, i \in [N+1,2N]\f$ represents the positive amount of assets \f$i-N\f$ bought  at date \f$t\f$
///  - \f$ X^i_t, i \in [2N+1,3N]\f$ represents the positive amount of assets \f$i-2N\f$ sold  at date \f$t\f$
///  - \f$ X^{3N+1}_t\f$ is the positive amount invested in the non risky asset at date \f$t\f$ before rebalancing
/// \f$ c \f$ represents the transaction cost paid at sell or buy orders
/// The flow equations are
/// - \f$ X^i_t = (1+a^i_t) (X^i_{t-dt} - X^{i+N}_{t-dt} + X^{i+2N}_{t-dt}) \f$
/// - \f$ X^{3N+1}_t =  X^{3N+1}_{t-dt} + (1-c)  X^{i+N}_{t-dt}  -(1+c) X^{i+2N}_{t-dt} \f$
/// The cost function to minimize
/// \f$ E[ \sum_{t>0} ( -  \sum_i (a^i_t (X^i_{t-dt}-X^{i+N}_{t-dt} +X^{i+2N}_{t-dt})  + c X^{i+N}_{t-dt} + c X^{i+2N}_{t_dt}) ) ] \f$

template< class Simulator>
class OptimizerPortfolio : public StOpt::OptimizerSDDPBase
{
private:

    double m_commission ; ///< transaction cost
    int m_nbAsset; ///< number of risky assets
    Eigen::ArrayXd m_mu ; ///< average yield of assets between two time steps
    Eigen::ArrayXv m_sig; /// standard deviation of yield between two time steps
    Eigen::ArrayXd m_initState; ///< Initial states (size nbAsset +1 for the non risk asset) used to get an admissible state
    std::shared_ptr< Simulator> m_simulatorBackward ; // for backward simulations
    std::shared_ptr< Simulator> m_simulatorForward ; // for forward simulations

public:

    /// \brief Constructor
    /// \param p_commission          defines the, transition cost
    /// \param p_mu                  average yield of the risky assets
    /// \param p_sig                 standard deviation of the yield of the risky assets
    /// \param p_initState           defines an admissible state that will be kept for fist backward iteration
    /// \param p_simulatorBackward   backward  simulator
    /// \param p_simulatorForward    Forward simulator
    OptimizerPortfolio(const double &p_commission, const Eigen::ArrayXd &p_mu, const Eigen::ArrayXd &p_vol ,
                       const Eigen::ArrayXd &p_initState,
                       const std::shared_ptr< Simulator> &p_simulatorBackward, const std::shared_ptr< Simulator> &p_simulatorForward):
        m_commission(p_commission), m_nbAsset(p_nbAsset), m_initState(p_initState), m_simulatorBackward(p_simulatorBackward), m_simulatorForward(p_simulatorForward) {}



    /// \brief Optimize the LP during backward resolution
    /// \param p_linCut            cuts used for the PL   (Benders for the Bellman value at the end of the time step)
    /// \param p_aState            Store the state, and 0, 0 values
    /// \param p_particle          Here no regression , so empty array
    /// \param p_isample            sample number for independent uncertainties
    /// \return  a vector with the optimal value and the derivatives if the function value with respect to each state (here the stocks)
    Eigen::ArrayXd oneStepBackward(const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut, const std::tuple< std::shared_ptr<Eigen::ArrayXd>, int  , int > &p_aState,
                                   const Eigen::ArrayXd &p_particle, const int &p_isample) const
    {
        //std::cout << "numéro simu"<<std::get<1>(p_aState)<< std::endl <<"tirageratio=" << p_particle << std::endl;
        AddConstraintOptimizerPortfolioopti constraints(std::get<1>(p_aState));
        Eigen::ArrayXd stateFollowing(3 * m_nbassets + 1);
        Eigen::ArrayXd  valueAndDerivatives(m_nbassets + 1);
        double gain ;
        createAndSolveLP(p_linCut, *std::get<0>(p_aState), ratiocurrentpricepreviouspricesimu, constraints, valueAndDerivatives, stateFollowing, gain);
        return valueAndDerivatives;
    }

    /// \brief Optimize the LP during forward resolution
    /// \param p_aParticle         a particle in simulation part to get back conditional cuts
    /// \param p_state             Store the state, the particle number used in optimization (if conditional cuts) and mesh number associated to the particle. As an input it constraints the current state
    /// \param p_stateToStore      For backward resolution we need to store \f$ (S_t,A_{t-1},D_{t-1}) \f$  where p_state in output is \f$ (S_t,A_{t},D_{t}) \f$
    /// \param p_linCut            cuts used for the PL   (Benders for the Bellman value at the end of the time step)
    /// \param p_isample            sample number for independent uncertainties
    double  oneStepForward(const Eigen::ArrayXd &p_aParticle, Eigen::ArrayXd &p_state, Eigen::ArrayXd &p_stateToStore ,
                           const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut, const int &p_isample) const
    {
        // optimizer constraints
        //std::cout << "one step forward" << "date=" << m_date << std::endl ;
        Eigen::ArrayXd ratiocurrentpricepreviouspricesimu = p_aParticle;
        AddConstraintOptimizerPortfoliosimu constraints(ratiocurrentpricepreviouspricesimu);
        Eigen::ArrayXd stateFollowing(3 * m_nbassets + 1);
        Eigen::ArrayXd  valueAndDerivatives(m_nbassets + 1);
        double gain ;
        createAndSolveLP(p_linCut, p_state, ratiocurrentpricepreviouspricesimu, constraints, valueAndDerivatives, stateFollowing, gain);
        p_stateToStore  = stateFollowing ;
        p_state = stateFollowing;
        return gain ;
    }

    template< class TConstraint>
    void createAndSolveLP(const std::unique_ptr< StOpt::SDDPCutBase > &p_linCut, const Eigen::ArrayXd &p_state, const Eigen::ArrayXd p_ratiocurrentpricepreviouspricesimu, const TConstraint &p_constraints, Eigen::ArrayXd &p_valueAndDerivatives, Eigen::ArrayXd &p_stateFollowing, double &p_gain) const
    {
        // optimizer matrix constraints
        Eigen::ArrayXi rows(m_nbassets + 1); // row position
        Eigen::ArrayXi columns(m_nbassets + 1) ; // columns  position
        Eigen::ArrayXd elements(m_nbassets + 1) ; // constraints matrix values
        // bounds associated to matrix constraints
        Eigen::ArrayXd lowBoundConst(m_nbassets + 1);
        Eigen::ArrayXd upperBoundConst(m_nbassets + 1) ;

        // bounds on values
        Eigen::ArrayXd lowBound(3 * m_nbassets + 2);
        Eigen::ArrayXd upperBound(3 * m_nbassets + 2);

        lowBound = Eigen::ArrayXd::Zero(3 * m_nbassets + 2);
        upperBound = Eigen::ArrayXd::Constant(3 * m_nbassets + 2, StOpt::infty);
        lowBound(3 * m_nbassets + 1) = -StOpt::infty;

        lowBoundConst(m_nbassets) = p_state(3 * m_nbassets); //Init
        upperBoundConst(m_nbassets) = p_state(3 * m_nbassets);

        Eigen::ArrayXd objFunc = Eigen::ArrayXd::Zero(3 * m_nbassets + 2);

        for (int i_asset = 0; i_asset < m_nbassets; i_asset++)
        {
            columns(i_asset) = i_asset;
            rows(i_asset) = i_asset;
            elements(i_asset) = p_ratiocurrentpricepreviouspricesimu(i_asset);
            lowBoundConst(i_asset) = p_state(i_asset) + p_state(i_asset + m_nbassets) - p_state(i_asset + 2 * m_nbassets);
            upperBoundConst(i_asset) = p_state(i_asset) + p_state(i_asset + m_nbassets) - p_state(i_asset + 2 * m_nbassets);
            lowBoundConst(m_nbassets) += -(1 + m_commission) * p_state(i_asset + m_nbassets) + (1 - m_commission) * p_state(i_asset + 2 * m_nbassets);
            objFunc(i_asset) = 1 - 1 / elements(i_asset);
            objFunc(i_asset + m_nbassets) = m_commission;
            objFunc(i_asset + 2 * m_nbassets) = m_commission;
        }
        objFunc(3 * m_nbassets + 1) = 1 ;
        elements(m_nbassets) = 1;
        rows(m_nbassets) = m_nbassets;
        columns(m_nbassets) = 3 * m_nbassets;
        //	std::cout << "row" <<rows << std::endl ;
        //	std::cout << "col" << columns << std::endl ;
        //	std::cout << "elements" << elements << std::endl ;
        //	std::cout << "lowBoundConst" << lowBoundConst << std::endl ;
        //	std::cout << "upperBoundConst" << upperBoundConst << std::endl << std::endl ;
        //  add cuts for bellman value to constraints
        p_constraints.addConstraints(p_linCut, m_nbassets, rows, columns, elements, lowBoundConst, upperBoundConst);
        // Rajouter ajout contraintes + recupération des dérivées et fonction objectif

        //  model
        ClpSimplex  model;

#ifdef NDEBUG
        model.setLogLevel(0);
#endif
        //	std::cout << "avec coupe" << std::endl ;
        //	std::cout << "row" <<rows << std::endl ;
        //	std::cout << "col" << columns << std::endl ;
        //	std::cout << "elements" << elements << std::endl ;
        //	std::cout << "lowBoundConst" << lowBoundConst << std::endl ;
        //	std::cout << "upperBoundConst" << upperBoundConst << std::endl ;
        //	std::cout << "lowBoundConst" << lowBoundConst.size() << ";" << upperBoundConst.size() << ";" << rows.size() << ";" << columns.size() << ";" << elements.size()<< std::endl ;
        model.loadProblem(CoinPackedMatrix(false, rows.data(), columns.data(), elements.data(), elements.size()), lowBound.data(), upperBound.data(), objFunc.data(), lowBoundConst.data(), upperBoundConst.data());
        model.setOptimizationDirection(1) ; // minimize


        ClpSolve solvectl;
        solvectl.setSolveType(ClpSolve::usePrimal);
        solvectl.setSolveType(ClpSolve::useDual);
        solvectl.setPresolveType(ClpSolve::presolveOn);
        model.initialSolve(solvectl);

        bool modelSolved = model.isProvenOptimal();

        if (!modelSolved)
        {
            std::cout << "[problemLP::solveModel] : Warning Linear program could not be solved optimally somehow\n";
            //		std::cout << "etat" << p_state << std::endl ;
            //		std::cout << "ratio prix" << p_ratiocurrentpricepreviouspricesimu << std::endl ;
            //std::string cheminmps= "/home/i11677/Portfoliomanagementsddp/ModelPortfoliomanagementpasbon.mps";
            //model.writeMps(cheminmps.c_str());
            //		throw std::exception() ;
        }
        //else
        //{	//std::string cheminmps= "/home/i11677/Portfoliomanagementsddp/ModelPortfoliomanagementbon.mps";
        //model.writeMps(cheminmps.c_str());
        //}

        // optimal value
        p_valueAndDerivatives(0) = model.objectiveValue();

        // duals
        double *dual = model.dualRowSolution();
        for (int i_asset = 0; i_asset < m_nbassets; i_asset++)
        {
            p_valueAndDerivatives(1 + i_asset) = dual[i_asset];
        }

        // primal
        double *columnPrimal = model.primalColumnSolution();

        for (int isto = 0 ; isto < (3 * m_nbassets + 1) ; ++isto)
        {
            p_stateFollowing(isto) = columnPrimal[isto];
        }

        p_gain = model.objectiveValue() - columnPrimal[(3 * m_nbassets + 1)];

    }

    /// \brief get the backward simulator back
    std::shared_ptr< StOpt::SimulatorSDDPBase > getSimulatorBackward() const
    {
        return m_simulatorBackward ;
    }

    /// \brief get the forward simulator back
    std::shared_ptr< StOpt::SimulatorSDDPBase > getSimulatorForward() const
    {
        return m_simulatorForward ;
    }

    /// \brief get back state size
    inline int getStateSize() const
    {
        return m_nbassets;
    }

    /// brief update the optimizer for new date : useless here because coefficents doesn't depend on time
    void updateDates(const double &p_date, const double &p_dateNext)
    {  }

    std::shared_ptr<Eigen::ArrayXd> oneAdmissibleState(const double &date)
    {
        std::shared_ptr<Eigen::ArrayXd> toRet(new Eigen::ArrayXd(m_init_state)); //strategy buying/selling to keep the initial money
        return toRet;
    }
};


#endif
